/*
 * Copyright (c) 2004 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: sasl-common.c,v 1.12 2005/03/24 19:55:21 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/rpool.h"
#include "sm/string.h"
#include "sm/net.h"
#include "sm/sasl.h"
#include "sm/io.h"

#if SM_USE_SASL

static sasl_callback_t srvcallbacks[] =
{
#if 0
	{	SASL_CB_VERIFYFILE,	&safesaslfile,	NULL	},
	{	SASL_CB_PROXY_POLICY,	&proxy_policy,	NULL	},
#endif /* 0 */
	{	SASL_CB_LIST_END,	NULL,		NULL	}
};

/*
**  SM_SASLVERSIONOK -- check Cyrus SASL version
**
**	Parameters:
**		none
**
**	Returns:
**		usual sm_error code.
*/

sm_ret_T
sm_saslversionok(void)
{
	int version_major, version_minor, version_step;

	sasl_version_info(NULL, NULL, &version_major, &version_minor,
			&version_step, NULL);
	if (SASL_VERSION_MAJOR == version_major &&
	    SASL_VERSION_MINOR == version_minor &&
	    SASL_VERSION_STEP == version_step)
		return SM_SUCCESS;
	else
		return sm_error_perm(SM_EM_SASL, SM_E_VER_MIX);
}

/*
**  SM_SASLVERSIONPRT -- show Cyrus SASL version
**
**	Parameters:
**		fp -- file for output
**
**	Returns:
**		usual sm_error code.
*/

sm_ret_T
sm_saslversionprt(sm_file_T *fp)
{
	const char *implementation, *version_string;
	int version_major, version_minor, version_step, version_patch;

	sasl_version_info(&implementation, &version_string,
		&version_major, &version_minor,
		&version_step, &version_patch);

	sm_io_fprintf(fp,
		"Cyrus SASL versions:\n"
		"compiled against: version=%d.%d.%d\n"
		"linked against:   version=%d.%d.%d\n"
#if 0
		"implementation=%s, version=%s, patch=%d\n"
#endif /* 0 */
		, SASL_VERSION_MAJOR, SASL_VERSION_MINOR, SASL_VERSION_STEP
		, version_major, version_minor, version_step
#if 0
		, implementation, version_string, version_patch
#endif /* 0 */
		);
	return sm_saslversionok();
}

#if 0
/*
**  In order to ensure that storage leaks are tracked, and to prevent
**  conflicts between the sm_heap package and sasl, we tell sasl to
**  use the following heap allocation functions.  Unfortunately,
**  the sasl package incorrectly specifies the size of a block
**  using ulong: for portability, it should be size_t.
*/

void *sm_sasl_malloc(ulong);
static void *sm_sasl_calloc(ulong, ulong);
static void *sm_sasl_realloc(void *, ulong);
void sm_sasl_free(void *);

/*
**  SM_SASL_MALLOC -- malloc() for SASL
**
**	Parameters:
**		size -- size of requested memory.
**
**	Returns:
**		pointer to memory.
*/

void *
sm_sasl_malloc(ulong size)
{
	return sm_malloc((size_t) size);
}

/*
**  SM_SASL_CALLOC -- calloc() for SASL
**
**	Parameters:
**		nelem -- number of elements.
**		elemsize -- size of each element.
**
**	Returns:
**		pointer to memory.
**
**	Notice:
**		this isn't currently used by SASL.
*/

static void *
sm_sasl_calloc(ulong nelem, ulong elemsize)
{
	size_t size;
	void *p;

	size = (size_t) nelem * (size_t) elemsize;
	p = sm_malloc(size);
	if (p == NULL)
		return NULL;
	memset(p, '\0', size);
	return p;
}

/*
**  SM_SASL_REALLOC -- realloc() for SASL
**
**	Parameters:
**		p -- pointer to old memory.
**		size -- size of requested memory.
**
**	Returns:
**		pointer to new memory.
*/

static void *
sm_sasl_realloc(void *o, ulong size)
{
	return sm_realloc(o, (size_t) size);
}

/*
**  SM_SASL_FREE -- free() for SASL
**
**	Parameters:
**		p -- pointer to free.
**
**	Returns:
**		none
*/

void
sm_sasl_free(void *p)
{
	sm_free(p);
}
#endif /* 0 */

/*
**  SM_SASL_INIT -- initialize SASL
**
**	Parameters:
**		server -- server side?
**
**	Returns:
**		usual return code
*/

sm_ret_T
sm_sasl_init(bool server, uint32_t flags, sm_log_ctx_P lctx, sm_sasl_ctx_P *psasl_ctx)
{
	sm_ret_T ret;
	sm_sasl_ctx_P sasl_ctx;

	SM_REQUIRE(psasl_ctx != NULL);

	ret = SM_SUCCESS;
	*psasl_ctx = NULL;
	sasl_ctx = sm_zalloc(sizeof(*sasl_ctx));
	if (sasl_ctx == NULL)
		return sm_err_temp(ENOMEM);
	sasl_ctx->sm_sasl_flags = flags;
	sasl_ctx->sm_sasl_lctx = lctx;
#if 0
	sasl_set_alloc(sm_sasl_malloc, sm_sasl_calloc,
		       sm_sasl_realloc, sm_sasl_free);
#endif /* 0 */
	if (server)
	{
#if 1
		ret = sasl_server_init(srvcallbacks, "sendmail");
#else /* 0 */
		ret = sasl_server_init(NULL, "sendmail");
#endif /* 0 */
	}
	*psasl_ctx = sasl_ctx;
	return ret;
}

/*
**  ITEMINLIST -- does item appear in list?
**
**	Check whether item appears in list (which must be separated by a
**	character in delim) as a "word", i.e. it must appear at the begin
**	of the list or after a space, and it must end with a space or the
**	end of the list.
**
**	Parameters:
**		item -- item to search.
**		list -- list of items.
**		delim -- list of delimiters.
**
**	Returns:
**		pointer to occurrence (NULL if not found).
*/

char *
iteminlist(const char *item, char *list, const char *delim)
{
	char *s;
	int len;

	if (list == NULL || *list == '\0')
		return NULL;
	if (item == NULL || *item == '\0')
		return NULL;
	s = list;
	len = strlen(item);
	while (s != NULL && *s != '\0')
	{
		if (strncasecmp(s, item, len) == 0 &&
		    (s[len] == '\0' || strchr(delim, s[len]) != NULL))
			return s;
		s = strpbrk(s, delim);
		if (s != NULL)
			while (*++s == ' ')
				continue;
	}
	return NULL;
}

#if 0
/*
**  REMOVEMECH -- remove item [rem] from list [list]
**
**	Parameters:
**		rem -- item to remove
**		list -- list of items
**		rpool -- resource pool from which result is allocated.
**
**	Returns:
**		pointer to new list (NULL in case of error).
*/

char *
removemech(char *rem, char *list, SM_RPOOL_T *rpool)
{
	char *ret;
	char *needle;
	int len;

	if (list == NULL)
		return NULL;
	if (rem == NULL || *rem == '\0')
	{
		/* take out what? */
		return NULL;
	}

	/* find the item in the list */
	if ((needle = iteminlist(rem, list, " ")) == NULL)
	{
		/* not in there: return original */
		return list;
	}

	/* length of string without rem */
	len = strlen(list) - strlen(rem);
	if (len <= 0)
	{
		ret = (char *) sm_rpool_malloc_x(rpool, 1);
		*ret = '\0';
		return ret;
	}
	ret = (char *) sm_rpool_malloc_x(rpool, len);
	sm_memset(ret, '\0', len);

	/* copy from start to removed item */
	sm_memcpy(ret, list, needle - list);

	/* length of rest of string past removed item */
	len = strlen(needle) - strlen(rem) - 1;
	if (len > 0)
	{
		/* not last item -- copy into string */
		sm_memcpy(ret + (needle - list),
		       list + (needle - list) + strlen(rem) + 1,
		       len);
	}
	else
		ret[(needle - list) - 1] = '\0';
	return ret;
}
#endif /* 0 */

#if 0
/*
**  INTERSECT -- create the intersection between two lists
**
**	Parameters:
**		s1, s2 -- lists of items (separated by single blanks).
**		rpool -- resource pool from which result is allocated.
**
**	Returns:
**		the intersection of both lists.
*/

char *
sm_intersect(char *s1, char *s2, sm_rpool_P rpool)
{
	char *hr, *h1, *h, *res;
	int l1, l2, rl;

	if (s1 == NULL || s2 == NULL)	/* NULL string(s) -> NULL result */
		return NULL;
	l1 = strlen(s1);
	l2 = strlen(s2);
	rl = SM_MIN(l1, l2);
	res = (char *) sm_rpool_malloc(rpool, rl + 1);
	if (res == NULL)
		return NULL;
	*res = '\0';
	if (rl == 0)	/* at least one string empty? */
		return res;
	hr = res;
	h1 = s1;
	h = s1;

	/* walk through s1 */
	while (h != NULL && *h1 != '\0')
	{
		/* is there something after the current word? */
		if ((h = strchr(h1, ' ')) != NULL)
			*h = '\0';
		l1 = strlen(h1);

		/* does the current word appear in s2 ? */
		if (iteminlist(h1, s2, " ") != NULL)
		{
			/* add a blank if not first item */
			if (hr != res)
				*hr++ = ' ';

			/* copy the item */
			memcpy(hr, h1, l1);

			/* advance pointer in result list */
			hr += l1;
			*hr = '\0';
		}
		if (h != NULL)
		{
			/* there are more items */
			*h = ' ';
			h1 = h + 1;
		}
	}
	return res;
}
#endif /* 0 */

#if 0
/*
**  IPTOSTRING -- create string for SASL_IP*PORT property
**		(borrowed from lib/iptostring.c in Cyrus-IMAP)
**
**	Parameters:
**		addr -- (pointer to) socket address
**		addrlen -- length of socket address
**		out -- output string (result)
**		outlen -- maximum length of output string
**
**	Returns:
**		true iff successful.
**
**	Side Effects:
**		creates output string if successful.
**		sets errno if unsuccessful.
*/

#ifndef NI_WITHSCOPEID
# define NI_WITHSCOPEID	0
#endif
#ifndef NI_MAXHOST
# define NI_MAXHOST	1025
#endif
#ifndef NI_MAXSERV
# define NI_MAXSERV	32
#endif

bool
iptostring(sockaddr_P addr, sockaddr_len_T addrlen, char *out, uint outlen)
{
	char hbuf[NI_MAXHOST], pbuf[NI_MAXSERV];

	if (addr == NULL || out == NULL)
	{
		errno = EINVAL;
		return false;
	}

#if NETINET6
	if (getnameinfo((struct sockaddr *) addr, addrlen,
			hbuf, sizeof hbuf, pbuf, sizeof pbuf,
			NI_NUMERICHOST | NI_WITHSCOPEID | NI_NUMERICSERV) != 0)
		return false;
#else /* NETINET6 */
	if (addr->sa.sa_family != AF_INET)
	{
		errno = EINVAL;
		return false;
	}
	if (strlcpy(hbuf, inet_ntoa(addr->sin.sin_addr), sizeof(hbuf))
	    >= sizeof(hbuf))
	{
		errno = ENOMEM;
		return false;
	}
	sm_snprintf(pbuf, sizeof pbuf, "%d", ntohs(addr->sin.sin_port));
#endif /* NETINET6 */

	if (outlen < strlen(hbuf) + strlen(pbuf) + 2)
	{
		errno = ENOMEM;
		return false;
	}
	sm_snprintf(out, outlen, "%s;%s", hbuf, pbuf);
	return true;
}
#endif /* 0 */

/*
**  SM_SASLMECHS -- get list of possible AUTH mechanisms
**
**	Parameters:
**		sasl_ctx -- sendmail SASL context
**		conn -- SASL connection (context)
**		mech_list -- list of allowed SASL mechanisms (output)
**
**	Returns:
**		number of mechs.
*/

sm_ret_T
sm_saslmechs(sm_sasl_ctx_P sasl_ctx, sasl_conn_t *conn, char **mech_list)
{
	int len, num, result;

	SM_REQUIRE(sasl_ctx != NULL);
	SM_REQUIRE(mech_list != NULL);
	SM_REQUIRE(conn != NULL);

	/* "user" is currently unused */
	result = sasl_listmech(conn, NULL, "", " ", "",
			(const char **)mech_list,
			(uint *)&len, &num);
#if 0
	sm_io_fprintf(smioerr,
		"AUTH: result=%d, num=%d, mech_list=%s\n",
		result, num, *mech_list);
#endif /* 0 */
	if (result != SASL_OK)
	{
#if 0
		sm_log_write(sasl_ctx->sm_sasl_lctx;
			SS_LCAT_SERVER, SS_LMOD_SERVER,
			SM_LOG_WARN, 9,
			"sev=WARN, listmech=%d, num=%d",
			result, num);
#endif /* 0 */
		num = 0;
	}
	if (num == 0)
	{
		*mech_list = NULL;	/* be paranoid... */
#if 0
		if (result == SASL_OK && LogLevel > 9)
			sm_syslog(LOG_WARNING, NOQID,
				  "AUTH warning: no mechanisms");
#endif /* 0 */
	}
	return num;
}

#endif /* SM_USE_SASL */
