/*
 * Copyright (c) 2002-2005 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

/*
**  SMTPS - Loopback communication module
**  Useful for performance testing (ignoring the protocol overhead) and
**  to compare smtps with the SMTP sinks in statethreads/examples/
*/

#include "sm/generic.h"
SM_RCSID("@(#)$Id: s2n.c,v 1.18 2005/03/15 19:56:13 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/memops.h"
#include "sm/reccom.h"
#include "statethreads/st.h"
#include "sm/rcbst.h"
#include "sm/fcntl.h"
#include "sm/unixsock.h"
#include "sm/stsock.h"
#include "sm/stthreads.h"
#include "smtps.h"
#include "s2q.h"

#include "sm/smar.h"

#if SSQ_DEBUG
# include <stdio.h>
# define SSQ_DPRINTF(x) do {			\
	time_t currt = st_time();		\
	sm_io_fprintf(smioerr, "%6ld: ", (long) currt);	\
	sm_io_fprintf x;				\
	} while (0)

#else /* SSQ_DEBUG */
# define SSQ_DPRINTF(x)
#endif /* SSQ_DEBUG */

/* just a debugging help: number of open requests */
static uint Open_rqs = 0;

/* XXX timeout to send an RCB to QMGR */
#define SS_RCB_SND_TO	SEC2USEC(2)

/*
**  SS_CLR_SESS_RQ -- clear session request
**
**	Parameters:
**		s2q_ctx -- S2Q context
**		ss_sess -- session context
**
**	Returns:
**		usual sm_error code
**
**	Side Effects:
**		frees entry in array if found
*/

sm_ret_T
ss_clr_sess_rq(s2q_ctx_P s2q_ctx, ss_sess_P ss_sess)
{
	sm_ret_T ret;
	int i;

	SM_REQUIRE(s2q_ctx != NULL);
	SM_IS_SS_SESS(ss_sess);
	i = ss_sess->ssse_s2q_idx;
	if (i == SSSE_S2Q_IDX_NONE)
		return SM_SUCCESS;

	if  (i >= 0 && i < s2q_ctx->s2q_maxrcbs
	     && s2q_ctx->s2q_sess[i] != NULL
	     && s2q_ctx->s2q_sids[i] != NULL
	     && SESSTA_EQ(s2q_ctx->s2q_sids[i], ss_sess->ssse_id))
	{
		s2q_ctx->s2q_sess[i] = NULL;
		s2q_ctx->s2q_sids[i] = NULL;
		ss_sess->ssse_s2q_idx = SSSE_S2Q_IDX_NONE;
		SM_ASSERT(Open_rqs > 0);
		--Open_rqs;
		ret = SM_SUCCESS;
	}
	else
	{
		SSQ_DPRINTF((smioerr, "sev=ERROR, func=ss_clr_sess_rq, sess=%p, id=%s, i=%d, open=%u\n",
			ss_sess, ss_sess->ssse_id, i, Open_rqs));
		ret = sm_error_perm(SM_EM_SMTPS, EINVAL);
	}
	return ret;
}

/*
**  SM_W4Q2S_REPLY -- wait for reply from QMGR/SMAR, decode it, and store
**	the data in the session context.
**
**	Parameters:
**		ss_sess -- session context
**		tmo -- timeout (seconds)
**		com_flag -- flag to set if communication error occurred
**
**	Returns:
**		<=0: usual sm_error code
**		>0: (SMTP) reply code
**
**	Side Effects:
**		may set ss_sess->ssse_wr
*/

sm_ret_T
sm_w4q2s_reply(ss_sess_P ss_sess, uint tmo, s2q_ctx_P s2q_ctx)
{
	SM_IS_SS_SESS(ss_sess);
	return SM_SUCCESS;
}

/*
**  SM_S2Q_INIT -- initialize S2Q
**
**	Parameters:
**		ps2q_ctx -- (pointer to) S2Q context (output)
**		ss_ctx -- SMTP Server context
**		sockname -- name of socket to use for connection
**		wait4srv -- time to wait for server (s)
**		max_threads -- maximum number of threads
**		smtps_id -- id of SMTPS
**		maxrcbs -- maximum number of outstanding requests (RCBs)
**		notify -- notify server
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
sm_s2q_init(s2q_ctx_P *ps2q_ctx, ss_ctx_P ss_ctx, const char *sockname, uint wait4srv, uint max_threads, int smtps_id, uint maxrcbs, uint notify)
{
	sm_ret_T ret;
	sm_rcb_P rcb;
	size_t n;
	s2q_ctx_P s2q_ctx;

	SM_REQUIRE(ps2q_ctx != NULL);
	s2q_ctx = (s2q_ctx_P) sm_zalloc(sizeof(*s2q_ctx));
	if (s2q_ctx == NULL)
		goto errnomem;
	rcb = NULL;
	s2q_ctx->s2q_fd = INVALID_NETFD;
	s2q_ctx->s2q_smtps_id = smtps_id;
	s2q_ctx->s2q_wr_mutex = st_mutex_new();
	if (s2q_ctx->s2q_wr_mutex == NULL)
	{
		ret = sm_error_temp(SM_EM_SMTPS, errno);
		goto error;
	}

#if 0
	/* set this? LOOPBACK pipe? */
	s2q_ctx->s2q_fd
#endif /* 0 */

	s2q_ctx->s2q_maxrcbs = maxrcbs;
	s2q_ctx->s2q_currcbs = 0;
	n = maxrcbs * sizeof(*s2q_ctx->s2q_sids);
	s2q_ctx->s2q_sids = (sessta_id_P *) sm_zalloc(n);
	if (s2q_ctx->s2q_sids == NULL)
		goto errnomem;
	n = maxrcbs * sizeof(*s2q_ctx->s2q_sess);
	s2q_ctx->s2q_sess = (ss_sess_P *) sm_zalloc(n);
	if (s2q_ctx->s2q_sess == NULL)
		goto errnomem;
	s2q_ctx->s2q_status = S2Q_ST_OK;
	*ps2q_ctx = s2q_ctx;
	return SM_SUCCESS;

  errnomem:
	ret = sm_error_temp(SM_EM_SMTPS, ENOMEM);
  error:
	if (s2q_ctx != NULL)
	{
		SM_FREE(s2q_ctx->s2q_sids);
		SM_FREE(s2q_ctx->s2q_sess);
		if (s2q_ctx->s2q_wr_mutex != NULL)
		{
			st_mutex_destroy(s2q_ctx->s2q_wr_mutex);
			s2q_ctx->s2q_wr_mutex = NULL;
		}
		SM_FREE(s2q_ctx);
	}
	return ret;
}

/*
**  SM_S2Q_STOP -- stop S2Q
**
**	Parameters:
**		s2q_ctx -- S2Q context
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
sm_s2q_stop(s2q_ctx_P s2q_ctx)
{
	sm_ret_T ret;

	SM_REQUIRE(s2q_ctx != NULL);
	if (s2q_ctx->s2q_fd != INVALID_NETFD)
	{
		ret = un_st_socket_close(s2q_ctx->s2q_fd);
		if (sm_is_err(ret))
			goto error;
		s2q_ctx->s2q_fd = INVALID_NETFD;
	}
	s2q_ctx->s2q_status = S2Q_ST_CLOSED;
	if (s2q_ctx->s2q_wr_mutex != NULL)
	{
		st_mutex_destroy(s2q_ctx->s2q_wr_mutex);
		s2q_ctx->s2q_wr_mutex = NULL;
	}
	SM_FREE(s2q_ctx->s2q_sids);
	SM_FREE(s2q_ctx->s2q_sess);

	/* free s2q_ctx? only if allocated above! */
	return SM_SUCCESS;

  error:
	return ret;
}

/*
need functions that are the counterpart for:

qmgr_smtps_open(IN smtps-info, OUT status):
RT_S2Q_NID: int smtps-id
XXX should this also transfer an initial status?

qmgr_session_open(IN connection-info, IN session-id, OUT status):
RT_S2Q_ID: int smtps-id
RT_S2Q_NSEID: str session-id
RT_S2Q_CLTIPV4/RT_S2Q_CLTIPV6: str client IP address

qmgr_session_status(IN connection-info, IN session-id, IN session-status, OUT status):
RT_S2Q_ID: int smtps-id
RT_S2Q_SEID: str session-id
XXX: ?

qmgr_trans_open(IN session-id, IN sender-env-info, IN trans-id, OUT status):
RT_S2Q_ID: int smtps-id
RT_S2Q_NTAID: str transaction id
RT_S2Q_MAIL: str mail from

qmgr_rcpt_add(IN trans-id, IN rcpt-env-info, OUT status):
RT_S2Q_ID: int smtps-id
RT_S2Q_TAID: str transaction id
RT_S2Q_RCPT_IDX: int rcpt idx (?)
RT_S2Q_RCPT: str rcpt to
the last two values can be repeated to transmit a list.

qmgr_cdb(IN trans-id, IN cdb-id, OUT status):
RT_S2Q_ID: int smtps-id
RT_S2Q_TAID: str transaction-id
RT_S2Q_CDBID: str cdb-id

qmgr_trans_close(IN trans-id, IN cdb-id, OUT status):
RT_S2Q_ID: int smtps-id
RT_S2Q_CTAID: str transaction-id
RT_S2Q_CDBID: str cdb-id
see docs about this

qmgr_trans_discard(IN trans-id, OUT status):
RT_S2Q_ID: int smtps-id
RT_S2Q_DTAID: str transaction-id

qmgr_session_close(IN session-id, OUT status):
RT_S2Q_ID: int smtps-id
RT_S2Q_CSEID: str session-id

qmgr_smtps_close(IN smtps-info, OUT status):
RT_S2Q_CLID: int smtps-id

*/

/*
**  SM_S2Q_NSEID -- new session
**
**	Parameters:
**		ss_sess -- session context
**		s2q_ctx -- S2Q context
**		smtps_id -- id of SMTPS
**		sid -- session id
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
sm_s2q_nseid(ss_sess_P ss_sess, s2q_ctx_P s2q_ctx, int smtps_id, sessta_id_P sid)
{
	return SM_SUCCESS;
}

/*
**  SM_S2Q_CSEID -- close session
**
**	Parameters:
**		ss_sess -- session context
**		s2q_ctx -- S2Q context
**		smtps_id -- id of SMTPS
**		sid -- session id
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
sm_s2q_cseid(ss_sess_P ss_sess, s2q_ctx_P s2q_ctx, int smtps_id, sessta_id_P sid)
{
	return SM_SUCCESS;
}

/*
**  SM_S2Q_NTAID -- new transaction
**
**	Parameters:
**		ss_sess -- session context
**		s2q_ctx -- S2Q context
**		smtps_id -- id of SMTPS
**		tid -- transaction id
**		mail -- mail from address
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
sm_s2q_ntaid(ss_sess_P ss_sess, s2q_ctx_P s2q_ctx, int smtps_id, sessta_id_P tid, sm_str_P mail)
{
	return SM_SUCCESS;
}

/*
**  SM_S2Q_DTAID -- discard transaction
**
**	Parameters:
**		ss_sess -- session context
**		s2q_ctx -- S2Q context
**		smtps_id -- id of SMTPS
**		tid -- transaction id
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
sm_s2q_dtaid(ss_sess_P ss_sess, s2q_ctx_P s2q_ctx, int smtps_id, sessta_id_P tid)
{
	return SM_SUCCESS;
}


/*
**  SM_S2Q_RCPTID -- new recipient
**
**	Parameters:
**		ss_sess -- session context
**		s2q_ctx -- S2Q context
**		smtps_id -- id of SMTPS
**		tid -- transaction id
**		rcpt_idx -- rcpt idx
**		rcpt -- rcpt to address
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
sm_s2q_rcptid(ss_sess_P ss_sess, s2q_ctx_P s2q_ctx, int smtps_id, sessta_id_P tid, int rcpt_idx, sm_str_P rcpt)
{
	SM_REQUIRE(s2q_ctx != NULL);
	SM_IS_SS_SESS(ss_sess);
	return SM_SUCCESS;
}

/*
**  SM_S2Q_CTAID -- close transaction
**
**	Parameters:
**		ss_sess -- session context
**		s2q_ctx -- S2Q context
**		smtps_id -- id of SMTPS
**		tid -- transaction id
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
sm_s2q_ctaid(ss_sess_P ss_sess, s2q_ctx_P s2q_ctx, int smtps_id, sessta_id_P tid)
{
	SM_REQUIRE(s2q_ctx != NULL);
	SM_IS_SS_SESS(ss_sess);
	return SM_SUCCESS;
}

/*
**  SM_S2A_RCPT -- send recipient to SMAR
**
**	Parameters:
**		ss_sess -- session context
**		s2a_ctx -- S2A context
**		smtps_id -- id of SMTPS
**		tid -- transaction id
**		rcpt_idx -- rcpt idx
**		rcpt -- rcpt to address
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
sm_s2a_rcptid(ss_sess_P ss_sess, s2q_ctx_P s2a_ctx, int smtps_id, sessta_id_P tid, int rcpt_idx, uint32_t lt, sm_str_P rcpt)
{
	SM_REQUIRE(s2a_ctx != NULL);
	SM_IS_SS_SESS(ss_sess);
	return SM_SUCCESS;
}

/*
**  SM_S2A_ADDR -- send address to SMAR
**
**	Parameters:
**		ss_sess -- session context
**		s2a_ctx -- S2A context
**		smtps_id -- id of SMTPS
**		tid -- transaction id
**		addr -- rcpt to address
**		what -- record type
**		ltype -- lookup type
**		lflags -- lookup flags
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
sm_s2a_addr(ss_sess_P ss_sess, s2q_ctx_P s2a_ctx, int smtps_id, sessta_id_P tid, sm_str_P addr, uint32_t what, uint32_t ltype, uint32_t lflags)
{
	SM_REQUIRE(s2a_ctx != NULL);
	SM_IS_SS_SESS(ss_sess);
	return SM_SUCCESS;
}

/*
**  SM_S2A_CLT -- send client address to SMAR
**
**	Parameters:
**		ss_sess -- session context
**		s2a_ctx -- S2A context
**		smtps_id -- id of SMTPS
**		sid -- session id
**		addr -- address
**		what -- record type
**		ltype -- lookup type
**		lflags -- lookup flags
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
sm_s2a_clt(ss_sess_P ss_sess, s2q_ctx_P s2a_ctx, int smtps_id, sessta_id_P sid, sm_str_P clt, uint32_t what, uint32_t ltype, uint32_t lflags)
{
	return SM_SUCCESS;
}
sm_ret_T
sm_s2a_tls(ss_sess_P ss_sess, s2q_ctx_P s2a_ctx, int smtps_id)
{
	return SM_SUCCESS;
}

#if SM_USE_PMILTER
sm_ret_T
sm_s2m_clt(ss_sess_P ss_sess, s2q_ctx_P s2m_ctx, int smtps_id, sessta_id_P sid, sm_str_P clt)
{
	return SM_SUCCESS;
}
sm_ret_T
sm_s2m_helo(ss_sess_P ss_sess, s2q_ctx_P s2m_ctx, int smtps_id, sessta_id_P sid, sm_str_P helo)
{
	return SM_SUCCESS;
}
sm_ret_T
sm_s2m_mail(ss_sess_P ss_sess, s2q_ctx_P s2m_ctx, int smtps_id, sessta_id_P sid, sessta_id_P tid, sm_str_P mail)
{
	return SM_SUCCESS;
}
sm_ret_T
sm_s2m_rcpt(ss_sess_P ss_sess, s2q_ctx_P s2m_ctx, int smtps_id, sessta_id_P sid, int rcpt_idx, sm_str_P rcpt)
{
	return SM_SUCCESS;
}
sm_ret_T
sm_s2m_data(ss_sess_P ss_sess, s2q_ctx_P s2m_ctx, int smtps_id, sessta_id_P sid)
{
	return SM_SUCCESS;
}
sm_ret_T
sm_s2m_body(ss_sess_P ss_sess, s2q_ctx_P s2m_ctx, int smtps_id, sessta_id_P sid, uchar *buf, size_t len, bool last)
{
	return SM_SUCCESS;
}
#endif /* SM_USE_PMILTER */
